var slice = Array.prototype.slice;
var extend = function (SuperClass, Class, prototype, members) {
  var descriptors = function (object) {
    var base = {}, descriptor;
    for (var key in object) {
      descriptor = Object.getOwnPropertyDescriptor(object, key);
      if (!('get' in descriptor) && !('set' in descriptor)) {
        descriptor.enumerable = false;
      }
      base[key] = descriptor;
    }
    return base;
  };
  if (SuperClass) {
    Class.__proto__ = SuperClass;
  }
  Object.defineProperty(Class, 'prototype', { value: Object.create(SuperClass === null ? null : SuperClass.prototype, descriptors(prototype)) });
  Object.defineProperty(Class.prototype, 'constructor', { value: Class });
  if (members) {
    Object.defineProperties(Class, descriptors(members));
  }
  return Class;
};
var expect = require('chai').expect;
describe('harmonizer', function () {
  it('should support anonymous classes', function () {
    var Animal = function () {
        function Animal() {
          var proto = Object.getPrototypeOf(Animal.prototype);
          if (proto !== null)
            proto.constructor.apply(this, arguments);
        }
        return extend(Object, Animal, {
          sayHi: function () {
            return 'Hi, I am a ' + this.type() + '.';
          }
        }, {
          getName: function () {
            return 'Animal';
          }
        });
      }();
    var Dog = function () {
        function Dog() {
          var proto = Object.getPrototypeOf(Dog.prototype);
          if (proto !== null)
            proto.constructor.apply(this, arguments);
        }
        return extend(Animal, Dog, {
          type: function () {
            return 'dog';
          },
          sayHi: function () {
            return Object.getPrototypeOf(Dog.prototype).sayHi.call(this) + ' WOOF!';
          }
        }, {
          getName: function () {
            return Object.getPrototypeOf(Dog).getName.call(this) + '/Dog';
          }
        });
      }();
    expect(new Dog().sayHi()).to.equal('Hi, I am a dog. WOOF!');
    expect(Dog.getName()).to.equal('Animal/Dog');
    var count = 0;
    var Cat = function () {
        function Cat() {
          var proto = Object.getPrototypeOf(Cat.prototype);
          if (proto !== null)
            proto.constructor.apply(this, arguments);
        }
        return extend(function () {
          count++;
          return Animal;
        }(), Cat, {});
      }();
    Cat;
    expect(count).to.equal(1);
  });
  it('should support super method calls', function () {
    function Animal() {
      var proto = Object.getPrototypeOf(Animal.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Animal, {
      sayHi: function () {
        return 'I am an animal.';
      },
      sayOther: function () {
        return 'WAT?!';
      }
    });
    function Horse() {
      var proto = Object.getPrototypeOf(Horse.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Animal, Horse, {
      sayHi: function () {
        return Object.getPrototypeOf(Horse.prototype).sayOther.call(this);
      },
      sayOther: function () {
        return 'I see dead objects.';
      }
    });
    expect(new Horse().sayHi()).to.equal('WAT?!');
  });
  it('should support super method calls', function () {
    function Animal() {
      var proto = Object.getPrototypeOf(Animal.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Animal, {
      sayHi: function () {
        return 'I am an animal.';
      },
      sayOther: function () {
        return 'WAT?!';
      }
    });
    function Horse() {
      var proto = Object.getPrototypeOf(Horse.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Animal, Horse, {
      sayHi: function () {
        return Object.getPrototypeOf(Horse.prototype).sayOther.call(this);
      },
      sayOther: function () {
        return 'I see dead objects.';
      }
    });
    expect(new Horse().sayHi()).to.equal('WAT?!');
  });
  it('should support class expressions', function () {
    var Person = function () {
        function Person() {
          var proto = Object.getPrototypeOf(Person.prototype);
          if (proto !== null)
            proto.constructor.apply(this, arguments);
        }
        return extend(Object, Person, {});
      }();
    expect(typeof Person).to.equal('function');
    expect(function () {
      return function () {
        function Person() {
          var proto = Object.getPrototypeOf(Person.prototype);
          if (proto !== null)
            proto.constructor.apply(this, arguments);
        }
        return extend(Object, Person, {});
      }();
    }().name).to.equal('Person');
    expect(typeof function () {
      function Class() {
        var proto = Object.getPrototypeOf(Class.prototype);
        if (proto !== null)
          proto.constructor.apply(this, arguments);
      }
      return extend(Object, Class, {});
    }()).to.equal('function');
  });
  it('should support extends', function () {
    function Animal() {
      var proto = Object.getPrototypeOf(Animal.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Animal, {
      sayHi: function () {
        return 'Hi, I am a ' + this.type() + '.';
      }
    });
    function Dog() {
      var proto = Object.getPrototypeOf(Dog.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Animal, Dog, {
      type: function () {
        return 'dog';
      },
      sayHi: function () {
        return Object.getPrototypeOf(Dog.prototype).sayHi.call(this) + ' WOOF!';
      }
    });
    expect(new Dog().sayHi()).to.equal('Hi, I am a dog. WOOF!');
  });
  it('should support constructors', function () {
    function Multiplier(n) {
      if (n === void 0)
        n = 1;
      this.n = n;
    }
    extend(Object, Multiplier, {
      multiply: function (n) {
        if (n === void 0)
          n = 1;
        return n * this.n;
      }
    });
    expect(new Multiplier().n).to.equal(1);
    expect(new Multiplier(6).n).to.equal(6);
    expect(new Multiplier().multiply()).to.equal(1);
    expect(new Multiplier(2).multiply(3)).to.equal(6);
  });
  it('should support method declarations', function () {
    function Person() {
      var proto = Object.getPrototypeOf(Person.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Person, {
      getName: function () {
        return this.firstName + ' ' + this.lastName;
      }
    });
    var me = new Person();
    me.firstName = 'Brian';
    me.lastName = 'Donovan';
    expect(me.getName()).to.equal('Brian Donovan');
  });
  it('should support empty classes', function () {
    function Foo() {
      var proto = Object.getPrototypeOf(Foo.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Foo, {});
    expect(new Foo().constructor).to.equal(Foo);
    expect(new Foo() instanceof Foo).to.be.ok;
  });
  it('should have non enumerable prototype methods', function () {
    function Point(x, y) {
      this.x = x;
      this.y = y;
    }
    extend(Object, Point, {
      toString: function () {
        return '(' + this.x + ', ' + this.y + ')';
      }
    });
    var point = new Point(1, 2);
    var keys = [];
    for (var key in point) {
      keys.push(key);
    }
    expect(point.toString()).to.equal('(1, 2)');
    expect(keys.sort()).to.eql([
      'x',
      'y'
    ]);
  });
  it('should support extending null', function () {
    function Obj() {
      var proto = Object.getPrototypeOf(Obj.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(null, Obj, {});
    expect(Obj.toString).to.equal(Function.toString);
    expect(new Obj().toString).to.equal(undefined);
  });
  it('should support setters and getters', function () {
    function Person(firstName, lastName) {
      this.firstName = firstName;
      this.lastName = lastName;
    }
    extend(Object, Person, {
      get name() {
        return this.firstName + ' ' + this.lastName;
      },
      set name(name) {
        var parts = name.split(' ');
        this.firstName = parts[0];
        this.lastName = parts[1];
      }
    });
    var mazer = new Person('Mazer', 'Rackham');
    expect(mazer.name).to.equal('Mazer Rackham');
    mazer.name = 'Ender Wiggin';
    expect(mazer.firstName).to.equal('Ender');
    expect(mazer.lastName).to.equal('Wiggin');
    var forLoopProperties = [];
    for (var key in mazer) {
      forLoopProperties.push(key);
    }
    expect(forLoopProperties.indexOf('name') > -1).to.be.ok;
  });
  it('should support implicit superclass', function () {
    function Obj() {
      Object.getPrototypeOf(Obj.prototype).constructor.call(this);
    }
    extend(Object, Obj, {});
    new Obj();
  });
  it('should support method declaration with arguments', function () {
    function Tripler() {
      var proto = Object.getPrototypeOf(Tripler.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Tripler, {
      triple: function (n) {
        return n * 3;
      }
    });
    expect(new Tripler().triple(2)).to.equal(6);
  });
  it('should have writable methods', function () {
    var value;
    function Foo() {
      var proto = Object.getPrototypeOf(Foo.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Foo, {
      foo: function () {
        value = 1;
      }
    });
    var foo = new Foo();
    foo.foo = function () {
      value = 2;
    };
    foo.foo();
    expect(value).to.equal(2);
  });
  it('should support rest params', function () {
    function Joiner(string) {
      this.string = string;
    }
    extend(Object, Joiner, {
      join: function () {
        var items = slice.call(arguments, 0);
        return items.join(this.string);
      }
    }, {
      join: function (string) {
        var items = slice.call(arguments, 1);
        var joiner = new this(string);
        return joiner.join.apply(joiner, items);
      }
    });
    function ArrayLike() {
      var items = slice.call(arguments, 0);
      var self = this;
      items.forEach(function (item, i) {
        self[i] = item;
      });
      this.length = items.length;
    }
    extend(Object, ArrayLike, {});
    var joiner = new Joiner(' & ');
    expect(joiner.join(4, 5, 6)).to.equal('4 & 5 & 6');
    expect(new ArrayLike('a', 'b')[1]).to.equal('b');
  });
  it('should support static methods', function () {
    function Tripler() {
      var proto = Object.getPrototypeOf(Tripler.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Tripler, {}, {
      triple: function (n) {
        if (n === void 0)
          n = 1;
        return n * 3;
      },
      toString: function () {
        return '3' + Object.getPrototypeOf(Tripler).toString.call(this) + '3';
      }
    });
    function MegaTripler() {
      var proto = Object.getPrototypeOf(MegaTripler.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Tripler, MegaTripler, {}, {
      triple: function (n) {
        if (n === void 0)
          n = 1;
        return Object.getPrototypeOf(MegaTripler).triple.call(this, n) * Object.getPrototypeOf(MegaTripler).triple.call(this, n);
      }
    });
    var tripler = new Tripler();
    expect(Tripler.triple()).to.equal(3);
    expect(Tripler.triple(2), 6);
    expect(tripler.triple, undefined);
    expect(Tripler.toString()).to.equal('3' + Object.toString.call(Tripler) + '3');
    var mega = new MegaTripler();
    expect(MegaTripler.triple(2)).to.equal(36);
    expect(mega.triple).to.equal(undefined);
    expect(MegaTripler.toString()).to.equal('3' + Object.toString.call(MegaTripler) + '3');
  });
  it('should support changing the __proto__', function () {
    var log = '';
    function Base() {
      var proto = Object.getPrototypeOf(Base.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, Base, {
      p: function () {
        log += '[Base]';
      }
    });
    function OtherBase() {
      var proto = Object.getPrototypeOf(OtherBase.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Object, OtherBase, {
      p: function () {
        log += '[OtherBase]';
      }
    });
    function Derived() {
      var proto = Object.getPrototypeOf(Derived.prototype);
      if (proto !== null)
        proto.constructor.apply(this, arguments);
    }
    extend(Base, Derived, {
      p: function () {
        log += '[Derived]';
        Object.getPrototypeOf(Derived.prototype).p.call(this);
        Derived.prototype.__proto__ = OtherBase.prototype;
        Object.getPrototypeOf(Derived.prototype).p.call(this);
      }
    });
    new Derived().p();
    expect(log).to.equal('[Derived][Base][OtherBase]');
  });
});